@@ -3,16 +3,13 @@
 // Note: we don't use any of the standard 'cargo bench', 'test::Bencher',
 // etc. because it's unstable at the time of writing.
 
-use std::env;
-use std::fs;
 use std::io::{self, Read, Write};
-use std::ops::Deref;
-use std::ops::DerefMut;
+use std::ops::{Deref, DerefMut};
 use std::sync::Arc;
 use std::time::{Duration, Instant};
+use std::{env, fs};
 
 use pki_types::{CertificateDer, PrivateKeyDer};
-
 use rustls::client::Resumption;
 #[cfg(all(not(feature = "ring"), feature = "aws_lc_rs"))]
 use rustls::crypto::aws_lc_rs as provider;
@@ -24,10 +21,10 @@ use rustls::crypto::ring as provider;
 use rustls::crypto::ring::{cipher_suite, Ticketer};
 use rustls::crypto::CryptoProvider;
 use rustls::server::{NoServerSessionStorage, ServerSessionMemoryCache, WebPkiClientVerifier};
-use rustls::RootCertStore;
-use rustls::{ClientConfig, ClientConnection};
-use rustls::{ConnectionCommon, SideData};
-use rustls::{ServerConfig, ServerConnection};
+use rustls::{
+    ClientConfig, ClientConnection, ConnectionCommon, RootCertStore, ServerConfig,
+    ServerConnection, SideData,
+};
 
 pub fn main() {
     let mut args = std::env::args();
@@ -43,24 +40,6 @@ fn duration_nanos(d: Duration) -> f64 {
     (d.as_secs() as f64) + f64::from(d.subsec_nanos()) / 1e9
 }
 
-fn _bench<Fsetup, Ftest, S>(count: usize, name: &'static str, f_setup: Fsetup, f_test: Ftest)
-where
-    Fsetup: Fn() -> S,
-    Ftest: Fn(S),
-{
-    let mut times = Vec::new();
-
-    for _ in 0..count {
-        let state = f_setup();
-        let start = Instant::now();
-        f_test(state);
-        times.push(duration_nanos(Instant::now().duration_since(start)));
-    }
-
-    println!("{}", name);
-    println!("{:?}", times);
-}
-
 fn time<F>(mut f: F) -> f64
 where
     F: FnMut(),
@@ -165,7 +144,8 @@ impl ResumptionParam {
 #[derive(PartialEq, Clone, Copy, Debug)]
 enum KeyType {
     Rsa,
-    Ecdsa,
+    EcdsaP256,
+    EcdsaP384,
     Ed25519,
 }
 
@@ -190,25 +170,19 @@ impl BenchmarkParam {
 }
 
 static ALL_BENCHMARKS: &[BenchmarkParam] = &[
-    #[cfg(feature = "tls12")]
+    #[cfg(all(feature = "tls12", not(feature = "fips")))]
     BenchmarkParam::new(
         KeyType::Rsa,
         cipher_suite::TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
         &rustls::version::TLS12,
     ),
-    #[cfg(feature = "tls12")]
+    #[cfg(all(feature = "tls12", not(feature = "fips")))]
     BenchmarkParam::new(
-        KeyType::Ecdsa,
+        KeyType::EcdsaP256,
         cipher_suite::TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256,
         &rustls::version::TLS12,
     ),
     #[cfg(feature = "tls12")]
-    BenchmarkParam::new(
-        KeyType::Rsa,
-        cipher_suite::TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256,
-        &rustls::version::TLS12,
-    ),
-    #[cfg(feature = "tls12")]
     BenchmarkParam::new(
         KeyType::Rsa,
         cipher_suite::TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,
@@ -222,16 +196,17 @@ static ALL_BENCHMARKS: &[BenchmarkParam] = &[
     ),
     #[cfg(feature = "tls12")]
     BenchmarkParam::new(
-        KeyType::Ecdsa,
+        KeyType::EcdsaP256,
         cipher_suite::TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,
         &rustls::version::TLS12,
     ),
     #[cfg(feature = "tls12")]
     BenchmarkParam::new(
-        KeyType::Ecdsa,
+        KeyType::EcdsaP384,
         cipher_suite::TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,
         &rustls::version::TLS12,
     ),
+    #[cfg(not(feature = "fips"))]
     BenchmarkParam::new(
         KeyType::Rsa,
         cipher_suite::TLS13_CHACHA20_POLY1305_SHA256,
@@ -248,7 +223,7 @@ static ALL_BENCHMARKS: &[BenchmarkParam] = &[
         &rustls::version::TLS13,
     ),
     BenchmarkParam::new(
-        KeyType::Ecdsa,
+        KeyType::EcdsaP256,
         cipher_suite::TLS13_AES_128_GCM_SHA256,
         &rustls::version::TLS13,
     ),
@@ -263,7 +238,8 @@ impl KeyType {
     fn path_for(&self, part: &str) -> String {
         match self {
             Self::Rsa => format!("test-ca/rsa/{}", part),
-            Self::Ecdsa => format!("test-ca/ecdsa/{}", part),
+            Self::EcdsaP256 => format!("test-ca/ecdsa-p256/{}", part),
+            Self::EcdsaP384 => format!("test-ca/ecdsa-p384/{}", part),
             Self::Ed25519 => format!("test-ca/eddsa/{}", part),
         }
     }
@@ -682,3 +658,7 @@ fn all_tests() {
         bench_handshake(test, ClientAuth::Yes, ResumptionParam::Tickets);
     }
 }
+
+#[cfg(not(target_env = "msvc"))]
+#[global_allocator]
+static GLOBAL: tikv_jemallocator::Jemalloc = tikv_jemallocator::Jemalloc;
