@@ -4,34 +4,31 @@
 // https://boringssl.googlesource.com/boringssl/+/master/ssl/test
 //
 
+use std::fmt::{Debug, Formatter};
+use std::io::{self, BufReader, Read, Write};
+use std::sync::Arc;
+use std::{env, fs, net, process, thread, time};
+
+use base64::prelude::{Engine, BASE64_STANDARD};
+use pki_types::{CertificateDer, PrivateKeyDer, ServerName, UnixTime};
 use rustls::client::danger::{HandshakeSignatureValid, ServerCertVerified, ServerCertVerifier};
 use rustls::client::{ClientConfig, ClientConnection, Resumption, WebPkiServerVerifier};
+#[cfg(all(not(feature = "ring"), feature = "aws_lc_rs"))]
+use rustls::crypto::aws_lc_rs as provider;
+#[cfg(feature = "ring")]
+use rustls::crypto::ring as provider;
 use rustls::crypto::{CryptoProvider, SupportedKxGroup};
 use rustls::internal::msgs::codec::Codec;
 use rustls::internal::msgs::persist::ServerSessionValue;
 use rustls::server::danger::{ClientCertVerified, ClientCertVerifier};
 use rustls::server::{ClientHello, ServerConfig, ServerConnection, WebPkiClientVerifier};
 use rustls::{
-    self, client, server, sign, version, AlertDescription, CertificateError, Connection,
+    client, server, sign, version, AlertDescription, CertificateError, Connection,
     DigitallySignedStruct, DistinguishedName, Error, InvalidMessage, NamedGroup, PeerIncompatible,
     PeerMisbehaved, ProtocolVersion, RootCertStore, Side, SignatureAlgorithm, SignatureScheme,
     SupportedProtocolVersion,
 };
 
-#[cfg(all(not(feature = "ring"), feature = "aws_lc_rs"))]
-use rustls::crypto::aws_lc_rs as provider;
-#[cfg(feature = "ring")]
-use rustls::crypto::ring as provider;
-
-use base64::prelude::{Engine, BASE64_STANDARD};
-use pki_types::{CertificateDer, PrivateKeyDer, ServerName, UnixTime};
-
-use std::fmt::{Debug, Formatter};
-use std::io::{self, BufReader, Read, Write};
-use std::sync::Arc;
-use std::time;
-use std::{env, fs, net, process, thread};
-
 static BOGO_NACK: i32 = 89;
 
 macro_rules! println_err(
@@ -43,11 +40,13 @@ macro_rules! println_err(
 #[derive(Debug)]
 struct Options {
     port: u16,
+    shim_id: u64,
     side: Side,
     max_fragment: Option<usize>,
     resumes: usize,
     verify_peer: bool,
     require_any_client_cert: bool,
+    root_hint_subjects: Vec<DistinguishedName>,
     offer_no_client_cas: bool,
     tickets: bool,
     resume_with_tickets_disabled: bool,
@@ -84,12 +83,14 @@ struct Options {
     expect_version: u16,
     resumption_delay: u32,
     queue_early_data_after_received_messages: Vec<usize>,
+    require_ems: bool,
 }
 
 impl Options {
     fn new() -> Self {
         Options {
             port: 0,
+            shim_id: 0,
             side: Side::Client,
             max_fragment: None,
             resumes: 0,
@@ -105,6 +106,7 @@ impl Options {
             shut_down_after_handshake: false,
             check_close_notify: false,
             require_any_client_cert: false,
+            root_hint_subjects: vec![],
             offer_no_client_cas: false,
             key_file: "".to_string(),
             cert_file: "".to_string(),
@@ -131,12 +133,14 @@ impl Options {
             expect_version: 0,
             resumption_delay: 0,
             queue_early_data_after_received_messages: vec![],
+            require_ems: false,
         }
     }
 
     fn version_allowed(&self, vers: ProtocolVersion) -> bool {
-        (self.min_version.is_none() || vers.get_u16() >= self.min_version.unwrap().get_u16())
-            && (self.max_version.is_none() || vers.get_u16() <= self.max_version.unwrap().get_u16())
+        (self.min_version.is_none() || u16::from(vers) >= u16::from(self.min_version.unwrap()))
+            && (self.max_version.is_none()
+                || u16::from(vers) <= u16::from(self.max_version.unwrap()))
     }
 
     fn tls13_supported(&self) -> bool {
@@ -202,16 +206,26 @@ fn split_protocols(protos: &str) -> Vec<String> {
     ret
 }
 
+fn decode_hex(hex: &str) -> Vec<u8> {
+    (0..hex.len())
+        .step_by(2)
+        .map(|i| u8::from_str_radix(&hex[i..i + 2], 16).unwrap())
+        .inspect(|x| println!("item {:?}", x))
+        .collect()
+}
+
 #[derive(Debug)]
 struct DummyClientAuth {
     mandatory: bool,
+    root_hint_subjects: Vec<DistinguishedName>,
     parent: Arc<dyn ClientCertVerifier>,
 }
 
 impl DummyClientAuth {
-    fn new(mandatory: bool) -> Self {
+    fn new(mandatory: bool, root_hint_subjects: Vec<DistinguishedName>) -> Self {
         Self {
             mandatory,
+            root_hint_subjects,
             parent: WebPkiClientVerifier::builder_with_provider(
                 load_root_certs(),
                 provider::default_provider().into(),
@@ -232,7 +246,7 @@ impl ClientCertVerifier for DummyClientAuth {
     }
 
     fn root_hint_subjects(&self) -> &[DistinguishedName] {
-        &[]
+        &self.root_hint_subjects
     }
 
     fn verify_client_cert(
@@ -397,6 +411,7 @@ fn lookup_scheme(scheme: u16) -> SignatureScheme {
         0x0601 => SignatureScheme::RSA_PKCS1_SHA512,
         0x0403 => SignatureScheme::ECDSA_NISTP256_SHA256,
         0x0503 => SignatureScheme::ECDSA_NISTP384_SHA384,
+        0x0603 => SignatureScheme::ECDSA_NISTP521_SHA512,
         0x0804 => SignatureScheme::RSA_PSS_SHA256,
         0x0805 => SignatureScheme::RSA_PSS_SHA384,
         0x0806 => SignatureScheme::RSA_PSS_SHA512,
@@ -484,7 +499,10 @@ impl server::StoresServerSessions for ServerCacheWithResumptionDelay {
 fn make_server_cfg(opts: &Options) -> Arc<ServerConfig> {
     let client_auth =
         if opts.verify_peer || opts.offer_no_client_cas || opts.require_any_client_cert {
-            Arc::new(DummyClientAuth::new(opts.require_any_client_cert))
+            Arc::new(DummyClientAuth::new(
+                opts.require_any_client_cert,
+                opts.root_hint_subjects.clone(),
+            ))
         } else {
             server::WebPkiClientVerifier::no_client_auth()
         };
@@ -517,6 +535,7 @@ fn make_server_cfg(opts: &Options) -> Arc<ServerConfig> {
     cfg.session_storage = ServerCacheWithResumptionDelay::new(opts.resumption_delay);
     cfg.max_fragment_size = opts.max_fragment;
     cfg.send_tls13_tickets = 1;
+    cfg.require_ems = opts.require_ems;
 
     if opts.use_signing_scheme > 0 {
         let scheme = lookup_scheme(opts.use_signing_scheme);
@@ -665,6 +684,7 @@ fn make_client_cfg(opts: &Options) -> Arc<ClientConfig> {
     cfg.resumption = Resumption::store(ClientCacheWithoutKxHints::new(opts.resumption_delay));
     cfg.enable_sni = opts.use_sni;
     cfg.max_fragment_size = opts.max_fragment;
+    cfg.require_ems = opts.require_ems;
 
     if !opts.protocols.is_empty() {
         cfg.alpn_protocols = opts
@@ -742,6 +762,10 @@ fn handle_err(err: Error) -> ! {
         Error::PeerMisbehaved(PeerMisbehaved::TooMuchEarlyDataReceived) => {
             quit(":TOO_MUCH_READ_EARLY_DATA:")
         }
+        Error::PeerMisbehaved(PeerMisbehaved::SignedHandshakeWithUnadvertisedSigScheme)
+        | Error::PeerMisbehaved(PeerMisbehaved::SignedKxWithWrongAlgorithm) => {
+            quit(":WRONG_SIGNATURE_TYPE:")
+        }
         Error::PeerMisbehaved(_) => quit(":PEER_MISBEHAVIOUR:"),
         Error::NoCertificatesPresented => quit(":NO_CERTS:"),
         Error::AlertReceived(AlertDescription::UnexpectedMessage) => quit(":BAD_ALERT:"),
@@ -828,6 +852,9 @@ fn exec(opts: &Options, mut sess: Connection, count: usize) {
     let mut sent_exporter = false;
     let mut quench_writes = false;
 
+    conn.write_all(&opts.shim_id.to_le_bytes())
+        .unwrap();
+
     loop {
         if !sent_message && (opts.queue_data || (opts.queue_data_on_resume && count > 0)) {
             if !opts
@@ -1004,6 +1031,9 @@ pub fn main() {
             "-port" => {
                 opts.port = args.remove(0).parse::<u16>().unwrap();
             }
+            "-shim-id" => {
+                opts.shim_id = args.remove(0).parse::<u64>().unwrap();
+            }
             "-server" => {
                 opts.side = Side::Server;
             }
@@ -1055,6 +1085,21 @@ pub fn main() {
                 let alg = args.remove(0).parse::<u16>().unwrap();
                 opts.use_signing_scheme = alg;
             }
+            "-use-client-ca-list" => {
+                match args.remove(0).as_ref() {
+                    "<EMPTY>" | "<NULL>" => {
+                        opts.root_hint_subjects = vec![];
+                    }
+                    list => {
+                        opts.root_hint_subjects = list.split(',')
+                            .map(|entry| DistinguishedName::from(decode_hex(entry)))
+                            .collect();
+                    }
+                }
+            }
+            "-verify-prefs" => {
+                lookup_scheme(args.remove(0).parse::<u16>().unwrap());
+            }
             "-max-cert-list" |
             "-expect-curve-id" |
             "-expect-resume-curve-id" |
@@ -1217,6 +1262,9 @@ pub fn main() {
                 opts.resumption_delay = args.remove(0).parse::<u32>().unwrap();
                 align_time();
             }
+            "-expect-extended-master-secret" => {
+                opts.require_ems = true;
+            }
 
             // defaults:
             "-enable-all-curves" |
@@ -1225,10 +1273,10 @@ pub fn main() {
             "-no-tls1" |
             "-no-ssl3" |
             "-handoff" |
+            "-ipv6" |
             "-decline-alpn" |
             "-expect-no-session" |
             "-expect-session-miss" |
-            "-expect-extended-master-secret" |
             "-expect-ticket-renewal" |
             "-enable-ocsp-stapling" |
             // internal openssl details:
@@ -1267,7 +1315,6 @@ pub fn main() {
             "-expect-early-data-info" |
             "-expect-cipher-aes" |
             "-retain-only-sha256-client-cert-initial" |
-            "-use-client-ca-list" |
             "-expect-draft-downgrade" |
             "-allow-unknown-alpn-protos" |
             "-on-initial-tls13-variant" |
@@ -1278,7 +1325,6 @@ pub fn main() {
             "-handshake-twice" |
             "-on-resume-verify-fail" |
             "-reverify-on-resume" |
-            "-verify-prefs" |
             "-no-op-extra-handshake" |
             "-expect-peer-cert-file" |
             "-no-rsa-pss-rsae-certs" |
