@@ -1,50 +1,46 @@
+use alloc::boxed::Box;
+use alloc::sync::Arc;
+use alloc::vec;
+use alloc::vec::Vec;
+
+use pki_types::ServerName;
+use subtle::ConstantTimeEq;
+
+use super::client_conn::ClientConnectionData;
+use super::hs::ClientContext;
 use crate::check::inappropriate_handshake_message;
-#[cfg(feature = "quic")]
-use crate::common_state::Protocol;
-#[cfg(feature = "secret_extraction")]
-use crate::common_state::Side;
-use crate::common_state::{CommonState, State};
+use crate::client::common::{ClientAuthDetails, ClientHelloDetails, ServerCertDetails};
+use crate::client::{hs, ClientConfig, ClientSessionStore};
+use crate::common_state::{CommonState, Protocol, Side, State};
 use crate::conn::ConnectionRandoms;
+use crate::crypto::ActiveKeyExchange;
 use crate::enums::{
     AlertDescription, ContentType, HandshakeType, ProtocolVersion, SignatureScheme,
 };
 use crate::error::{Error, InvalidMessage, PeerIncompatible, PeerMisbehaved};
 use crate::hash_hs::{HandshakeHash, HandshakeHashBuffer};
-use crate::kx;
 #[cfg(feature = "logging")]
 use crate::log::{debug, trace, warn};
 use crate::msgs::base::{Payload, PayloadU8};
 use crate::msgs::ccs::ChangeCipherSpecPayload;
-use crate::msgs::enums::ExtensionType;
-use crate::msgs::enums::KeyUpdateRequest;
-use crate::msgs::handshake::NewSessionTicketPayloadTLS13;
-use crate::msgs::handshake::{CertificateEntry, CertificatePayloadTLS13};
-use crate::msgs::handshake::{ClientExtension, ServerExtension};
-use crate::msgs::handshake::{HandshakeMessagePayload, HandshakePayload};
-use crate::msgs::handshake::{HasServerExtensions, ServerHelloPayload};
-use crate::msgs::handshake::{PresharedKeyIdentity, PresharedKeyOffer};
+use crate::msgs::enums::{ExtensionType, KeyUpdateRequest};
+use crate::msgs::handshake::{
+    CertificateEntry, CertificatePayloadTls13, ClientExtension, HandshakeMessagePayload,
+    HandshakePayload, HasServerExtensions, NewSessionTicketPayloadTls13, PresharedKeyIdentity,
+    PresharedKeyOffer, ServerExtension, ServerHelloPayload,
+};
 use crate::msgs::message::{Message, MessagePayload};
 use crate::msgs::persist;
-#[cfg(feature = "secret_extraction")]
+use crate::sign::{CertifiedKey, Signer};
 use crate::suites::PartiallyExtractedSecrets;
 use crate::tls13::key_schedule::{
     KeyScheduleEarly, KeyScheduleHandshake, KeySchedulePreHandshake, KeyScheduleTraffic,
 };
-use crate::tls13::Tls13CipherSuite;
+use crate::tls13::{
+    construct_client_verify_message, construct_server_verify_message, Tls13CipherSuite,
+};
 use crate::verify::{self, DigitallySignedStruct};
-use crate::{sign, KeyLog};
-
-use super::client_conn::ClientConnectionData;
-use super::hs::ClientContext;
-use crate::client::common::ServerCertDetails;
-use crate::client::common::{ClientAuthDetails, ClientHelloDetails};
-use crate::client::{hs, ClientConfig, ClientSessionStore, ServerName};
-
-use crate::ticketer::TimeBase;
-use ring::constant_time;
-
-use crate::sign::{CertifiedKey, Signer};
-use std::sync::Arc;
+use crate::{crypto, KeyLog};
 
 // Extensions we expect in plaintext in the ServerHello.
 static ALLOWED_PLAINTEXT_EXTS: &[ExtensionType] = &[
@@ -67,56 +63,69 @@ pub(super) fn handle_server_hello(
     cx: &mut ClientContext,
     server_hello: &ServerHelloPayload,
     mut resuming_session: Option<persist::Tls13ClientSessionValue>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     early_key_schedule: Option<KeyScheduleEarly>,
     hello: ClientHelloDetails,
-    our_key_share: kx::KeyExchange,
+    our_key_share: Box<dyn ActiveKeyExchange>,
     mut sent_tls13_fake_ccs: bool,
-) -> hs::NextStateOrError {
+) -> hs::NextStateOrError<'static> {
     validate_server_hello(cx.common, server_hello)?;
 
     let their_key_share = server_hello
-        .get_key_share()
+        .key_share()
         .ok_or_else(|| {
-            cx.common
-                .send_fatal_alert(AlertDescription::MissingExtension);
-            Error::PeerMisbehaved(PeerMisbehaved::MissingKeyShare)
+            cx.common.send_fatal_alert(
+                AlertDescription::MissingExtension,
+                PeerMisbehaved::MissingKeyShare,
+            )
         })?;
 
     if our_key_share.group() != their_key_share.group {
-        return Err(cx
-            .common
-            .illegal_param(PeerMisbehaved::WrongGroupForKeyShare));
+        return Err({
+            cx.common.send_fatal_alert(
+                AlertDescription::IllegalParameter,
+                PeerMisbehaved::WrongGroupForKeyShare,
+            )
+        });
     }
 
     let key_schedule_pre_handshake = if let (Some(selected_psk), Some(early_key_schedule)) =
-        (server_hello.get_psk_index(), early_key_schedule)
+        (server_hello.psk_index(), early_key_schedule)
     {
         if let Some(ref resuming) = resuming_session {
             let resuming_suite = match suite.can_resume_from(resuming.suite()) {
                 Some(resuming) => resuming,
                 None => {
-                    return Err(cx.common.illegal_param(
-                        PeerMisbehaved::ResumptionOfferedWithIncompatibleCipherSuite,
-                    ));
+                    return Err({
+                        cx.common.send_fatal_alert(
+                            AlertDescription::IllegalParameter,
+                            PeerMisbehaved::ResumptionOfferedWithIncompatibleCipherSuite,
+                        )
+                    });
                 }
             };
 
             // If the server varies the suite here, we will have encrypted early data with
             // the wrong suite.
             if cx.data.early_data.is_enabled() && resuming_suite != suite {
-                return Err(cx
-                    .common
-                    .illegal_param(PeerMisbehaved::EarlyDataOfferedWithVariedCipherSuite));
+                return Err({
+                    cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::EarlyDataOfferedWithVariedCipherSuite,
+                    )
+                });
             }
 
             if selected_psk != 0 {
-                return Err(cx
-                    .common
-                    .illegal_param(PeerMisbehaved::SelectedInvalidPsk));
+                return Err({
+                    cx.common.send_fatal_alert(
+                        AlertDescription::IllegalParameter,
+                        PeerMisbehaved::SelectedInvalidPsk,
+                    )
+                });
             }
 
             debug!("Resuming using PSK");
@@ -134,21 +143,21 @@ pub(super) fn handle_server_hello(
         KeySchedulePreHandshake::new(suite)
     };
 
-    let key_schedule = our_key_share.complete(&their_key_share.payload.0, |secret| {
-        Ok(key_schedule_pre_handshake.into_handshake(secret))
-    })?;
+    let shared_secret = our_key_share.complete(&their_key_share.payload.0)?;
+
+    let key_schedule = key_schedule_pre_handshake.into_handshake(shared_secret);
 
     // Remember what KX group the server liked for next time.
     config
         .resumption
         .store
-        .set_kx_hint(&server_name, their_key_share.group);
+        .set_kx_hint(server_name.clone(), their_key_share.group);
 
     // If we change keying when a subsequent handshake message is being joined,
     // the two halves will have different record layer protections.  Disallow this.
     cx.common.check_aligned_handshake()?;
 
-    let hash_at_client_recvd_server_hello = transcript.get_current_hash();
+    let hash_at_client_recvd_server_hello = transcript.current_hash();
     let key_schedule = key_schedule.derive_client_handshake_secrets(
         cx.data.early_data.is_enabled(),
         hash_at_client_recvd_server_hello,
@@ -177,9 +186,11 @@ fn validate_server_hello(
     server_hello: &ServerHelloPayload,
 ) -> Result<(), Error> {
     for ext in &server_hello.extensions {
-        if !ALLOWED_PLAINTEXT_EXTS.contains(&ext.get_type()) {
-            common.send_fatal_alert(AlertDescription::UnsupportedExtension);
-            return Err(PeerMisbehaved::UnexpectedCleartextExtension.into());
+        if !ALLOWED_PLAINTEXT_EXTS.contains(&ext.ext_type()) {
+            return Err(common.send_fatal_alert(
+                AlertDescription::UnsupportedExtension,
+                PeerMisbehaved::UnexpectedCleartextExtension,
+            ));
         }
     }
 
@@ -188,21 +199,24 @@ fn validate_server_hello(
 
 pub(super) fn initial_key_share(
     config: &ClientConfig,
-    server_name: &ServerName,
-) -> Result<kx::KeyExchange, Error> {
+    server_name: &ServerName<'_>,
+) -> Result<Box<dyn ActiveKeyExchange>, Error> {
     let group = config
         .resumption
         .store
         .kx_hint(server_name)
-        .and_then(|group| kx::KeyExchange::choose(group, &config.kx_groups))
+        .and_then(|group_name| config.find_kx_group(group_name))
         .unwrap_or_else(|| {
             config
+                .provider
                 .kx_groups
-                .first()
+                .iter()
+                .copied()
+                .next()
                 .expect("No kx groups configured")
         });
 
-    kx::KeyExchange::start(group).ok_or(Error::FailedToGetRandomBytes)
+    group.start()
 }
 
 /// This implements the horrifying TLS1.3 hack where PSK binders have a
@@ -214,12 +228,12 @@ pub(super) fn fill_in_psk_binder(
 ) -> KeyScheduleEarly {
     // We need to know the hash function of the suite we're trying to resume into.
     let suite = resuming.suite();
-    let suite_hash = suite.hash_algorithm();
+    let suite_hash = suite.common.hash_provider;
 
     // The binder is calculated over the clienthello, but doesn't include itself or its
     // length, or the length of its container.
-    let binder_plaintext = hmp.get_encoding_for_binder_signing();
-    let handshake_hash = transcript.get_hash_given(suite_hash, &binder_plaintext);
+    let binder_plaintext = hmp.encoding_for_binder_signing();
+    let handshake_hash = transcript.hash_given(suite_hash, &binder_plaintext);
 
     // Run a fake key_schedule to simulate what the server will do if it chooses
     // to resume.
@@ -261,8 +275,9 @@ pub(super) fn prepare_resumption(
     let obfuscated_ticket_age = resuming_session.obfuscated_ticket_age();
 
     let binder_len = resuming_suite
-        .hash_algorithm()
-        .output_len;
+        .common
+        .hash_provider
+        .output_len();
     let binder = vec![0u8; binder_len];
 
     let psk_identity =
@@ -283,7 +298,7 @@ pub(super) fn derive_early_traffic_secret(
     // For middlebox compatibility
     emit_fake_ccs(sent_tls13_fake_ccs, cx.common);
 
-    let client_hello_hash = transcript_buffer.get_hash_given(resuming_suite.hash_algorithm(), &[]);
+    let client_hello_hash = transcript_buffer.hash_given(resuming_suite.common.hash_provider, &[]);
     early_key_schedule.client_early_traffic_secret(
         &client_hello_hash,
         key_log,
@@ -301,7 +316,7 @@ pub(super) fn emit_fake_ccs(sent_tls13_fake_ccs: &mut bool, common: &mut CommonS
         return;
     }
 
-    if std::mem::replace(sent_tls13_fake_ccs, true) {
+    if core::mem::replace(sent_tls13_fake_ccs, true) {
         return;
     }
 
@@ -318,21 +333,27 @@ fn validate_encrypted_extensions(
     exts: &Vec<ServerExtension>,
 ) -> Result<(), Error> {
     if exts.has_duplicate_extension() {
-        common.send_fatal_alert(AlertDescription::DecodeError);
-        return Err(PeerMisbehaved::DuplicateEncryptedExtensions.into());
+        return Err(common.send_fatal_alert(
+            AlertDescription::DecodeError,
+            PeerMisbehaved::DuplicateEncryptedExtensions,
+        ));
     }
 
     if hello.server_sent_unsolicited_extensions(exts, &[]) {
-        common.send_fatal_alert(AlertDescription::UnsupportedExtension);
-        return Err(PeerMisbehaved::UnsolicitedEncryptedExtension.into());
+        return Err(common.send_fatal_alert(
+            AlertDescription::UnsupportedExtension,
+            PeerMisbehaved::UnsolicitedEncryptedExtension,
+        ));
     }
 
     for ext in exts {
-        if ALLOWED_PLAINTEXT_EXTS.contains(&ext.get_type())
-            || DISALLOWED_TLS13_EXTS.contains(&ext.get_type())
+        if ALLOWED_PLAINTEXT_EXTS.contains(&ext.ext_type())
+            || DISALLOWED_TLS13_EXTS.contains(&ext.ext_type())
         {
-            common.send_fatal_alert(AlertDescription::UnsupportedExtension);
-            return Err(PeerMisbehaved::DisallowedEncryptedExtension.into());
+            return Err(common.send_fatal_alert(
+                AlertDescription::UnsupportedExtension,
+                PeerMisbehaved::DisallowedEncryptedExtension,
+            ));
         }
     }
 
@@ -342,7 +363,7 @@ fn validate_encrypted_extensions(
 struct ExpectEncryptedExtensions {
     config: Arc<ClientConfig>,
     resuming_session: Option<persist::Tls13ClientSessionValue>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
@@ -351,7 +372,14 @@ struct ExpectEncryptedExtensions {
 }
 
 impl State<ClientConnectionData> for ExpectEncryptedExtensions {
-    fn handle(mut self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
         let exts = require_handshake_msg!(
             m,
             HandshakeType::EncryptedExtensions,
@@ -361,19 +389,16 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
         self.transcript.add_message(&m);
 
         validate_encrypted_extensions(cx.common, &self.hello, exts)?;
-        hs::process_alpn_protocol(cx.common, &self.config, exts.get_alpn_protocol())?;
+        hs::process_alpn_protocol(cx.common, &self.config, exts.alpn_protocol())?;
 
-        #[cfg(feature = "quic")]
-        {
-            // QUIC transport parameters
-            if cx.common.is_quic() {
-                match exts.get_quic_params_extension() {
-                    Some(params) => cx.common.quic.params = Some(params),
-                    None => {
-                        return Err(cx
-                            .common
-                            .missing_extension(PeerMisbehaved::MissingQuicTransportParameters));
-                    }
+        // QUIC transport parameters
+        if cx.common.is_quic() {
+            match exts.quic_params_extension() {
+                Some(params) => cx.common.quic.params = Some(params),
+                None => {
+                    return Err(cx
+                        .common
+                        .missing_extension(PeerMisbehaved::MissingQuicTransportParameters));
                 }
             }
         }
@@ -398,7 +423,7 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
             cx.common.peer_certificates = Some(
                 resuming_session
                     .server_cert_chain()
-                    .to_vec(),
+                    .clone(),
             );
 
             // We *don't* reverify the certificate chain here: resumption is a
@@ -427,29 +452,38 @@ impl State<ClientConnectionData> for ExpectEncryptedExtensions {
                 suite: self.suite,
                 transcript: self.transcript,
                 key_schedule: self.key_schedule,
-                may_send_sct_list: self.hello.server_may_send_sct_list(),
             }))
         }
     }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
 }
 
 struct ExpectCertificateOrCertReq {
     config: Arc<ClientConfig>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleHandshake,
-    may_send_sct_list: bool,
 }
 
 impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
-    fn handle(self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+    fn handle<'m>(
+        self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
         match m.payload {
             MessagePayload::Handshake {
                 parsed:
                     HandshakeMessagePayload {
-                        payload: HandshakePayload::CertificateTLS13(..),
+                        payload: HandshakePayload::CertificateTls13(..),
                         ..
                     },
                 ..
@@ -460,14 +494,13 @@ impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
                 suite: self.suite,
                 transcript: self.transcript,
                 key_schedule: self.key_schedule,
-                may_send_sct_list: self.may_send_sct_list,
                 client_auth: None,
             })
             .handle(cx, m),
             MessagePayload::Handshake {
                 parsed:
                     HandshakeMessagePayload {
-                        payload: HandshakePayload::CertificateRequestTLS13(..),
+                        payload: HandshakePayload::CertificateRequestTls13(..),
                         ..
                     },
                 ..
@@ -478,7 +511,6 @@ impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
                 suite: self.suite,
                 transcript: self.transcript,
                 key_schedule: self.key_schedule,
-                may_send_sct_list: self.may_send_sct_list,
             })
             .handle(cx, m),
             payload => Err(inappropriate_handshake_message(
@@ -491,6 +523,10 @@ impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
             )),
         }
     }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
 }
 
 // TLS1.3 version of CertificateRequest handling.  We then move to expecting the server
@@ -498,20 +534,26 @@ impl State<ClientConnectionData> for ExpectCertificateOrCertReq {
 // in TLS1.3.
 struct ExpectCertificateRequest {
     config: Arc<ClientConfig>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleHandshake,
-    may_send_sct_list: bool,
 }
 
 impl State<ClientConnectionData> for ExpectCertificateRequest {
-    fn handle(mut self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
         let certreq = &require_handshake_msg!(
             m,
             HandshakeType::CertificateRequest,
-            HandshakePayload::CertificateRequestTLS13
+            HandshakePayload::CertificateRequestTls13
         )?;
         self.transcript.add_message(&m);
         debug!("Got CertificateRequest {:?}", certreq);
@@ -522,32 +564,33 @@ impl State<ClientConnectionData> for ExpectCertificateRequest {
         // Must be empty during handshake.
         if !certreq.context.0.is_empty() {
             warn!("Server sent non-empty certreq context");
-            cx.common
-                .send_fatal_alert(AlertDescription::DecodeError);
-            return Err(InvalidMessage::InvalidCertRequest.into());
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::DecodeError,
+                InvalidMessage::InvalidCertRequest,
+            ));
         }
 
-        let tls13_sign_schemes = sign::supported_sign_tls13();
         let no_sigschemes = Vec::new();
         let compat_sigschemes = certreq
-            .get_sigalgs_extension()
+            .sigalgs_extension()
             .unwrap_or(&no_sigschemes)
             .iter()
             .cloned()
-            .filter(|scheme| tls13_sign_schemes.contains(scheme))
+            .filter(SignatureScheme::supported_in_tls13)
             .collect::<Vec<SignatureScheme>>();
 
         if compat_sigschemes.is_empty() {
-            cx.common
-                .send_fatal_alert(AlertDescription::HandshakeFailure);
-            return Err(PeerIncompatible::NoCertificateRequestSignatureSchemesInCommon.into());
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::HandshakeFailure,
+                PeerIncompatible::NoCertificateRequestSignatureSchemesInCommon,
+            ));
         }
 
         let client_auth = ClientAuthDetails::resolve(
             self.config
                 .client_auth_cert_resolver
                 .as_ref(),
-            certreq.get_authorities_extension(),
+            certreq.authorities_extension(),
             &compat_sigschemes,
             Some(certreq.context.0.clone()),
         );
@@ -559,66 +602,59 @@ impl State<ClientConnectionData> for ExpectCertificateRequest {
             suite: self.suite,
             transcript: self.transcript,
             key_schedule: self.key_schedule,
-            may_send_sct_list: self.may_send_sct_list,
             client_auth: Some(client_auth),
         }))
     }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
 }
 
 struct ExpectCertificate {
     config: Arc<ClientConfig>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleHandshake,
-    may_send_sct_list: bool,
     client_auth: Option<ClientAuthDetails>,
 }
 
 impl State<ClientConnectionData> for ExpectCertificate {
-    fn handle(mut self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
-        let cert_chain = require_handshake_msg!(
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
+        self.transcript.add_message(&m);
+        let cert_chain = require_handshake_msg_move!(
             m,
             HandshakeType::Certificate,
-            HandshakePayload::CertificateTLS13
+            HandshakePayload::CertificateTls13
         )?;
-        self.transcript.add_message(&m);
 
         // This is only non-empty for client auth.
         if !cert_chain.context.0.is_empty() {
-            warn!("certificate with non-empty context during handshake");
-            cx.common
-                .send_fatal_alert(AlertDescription::DecodeError);
-            return Err(InvalidMessage::InvalidCertRequest.into());
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::DecodeError,
+                InvalidMessage::InvalidCertRequest,
+            ));
         }
 
         if cert_chain.any_entry_has_duplicate_extension()
             || cert_chain.any_entry_has_unknown_extension()
         {
-            warn!("certificate chain contains unsolicited/unknown extension");
-            cx.common
-                .send_fatal_alert(AlertDescription::UnsupportedExtension);
-            return Err(PeerMisbehaved::BadCertChainExtensions.into());
-        }
-
-        let server_cert = ServerCertDetails::new(
-            cert_chain.convert(),
-            cert_chain.get_end_entity_ocsp(),
-            cert_chain
-                .get_end_entity_scts()
-                .map(|scts| scts.to_vec()),
-        );
-
-        if let Some(sct_list) = server_cert.scts.as_ref() {
-            if hs::sct_list_is_invalid(sct_list) {
-                return Err(PeerMisbehaved::InvalidSctList.into());
-            }
-
-            if !self.may_send_sct_list {
-                return Err(PeerMisbehaved::UnsolicitedSctList.into());
-            }
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::UnsupportedExtension,
+                PeerMisbehaved::BadCertChainExtensions,
+            ));
         }
+        let end_entity_ocsp = cert_chain.end_entity_ocsp();
+        let server_cert = ServerCertDetails::new(cert_chain.convert(), end_entity_ocsp);
 
         Ok(Box::new(ExpectCertificateVerify {
             config: self.config,
@@ -631,22 +667,33 @@ impl State<ClientConnectionData> for ExpectCertificate {
             client_auth: self.client_auth,
         }))
     }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
 }
 
 // --- TLS1.3 CertificateVerify ---
-struct ExpectCertificateVerify {
+struct ExpectCertificateVerify<'a> {
     config: Arc<ClientConfig>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleHandshake,
-    server_cert: ServerCertDetails,
+    server_cert: ServerCertDetails<'a>,
     client_auth: Option<ClientAuthDetails>,
 }
 
-impl State<ClientConnectionData> for ExpectCertificateVerify {
-    fn handle(mut self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+impl State<ClientConnectionData> for ExpectCertificateVerify<'_> {
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
         let cert_verify = require_handshake_msg!(
             m,
             HandshakeType::CertificateVerify,
@@ -661,7 +708,9 @@ impl State<ClientConnectionData> for ExpectCertificateVerify {
             .cert_chain
             .split_first()
             .ok_or(Error::NoCertificatesPresented)?;
-        let now = std::time::SystemTime::now();
+
+        let now = self.config.current_time()?;
+
         let cert_verified = self
             .config
             .verifier
@@ -669,7 +718,6 @@ impl State<ClientConnectionData> for ExpectCertificateVerify {
                 end_entity,
                 intermediates,
                 &self.server_name,
-                &mut self.server_cert.scts(),
                 &self.server_cert.ocsp_response,
                 now,
             )
@@ -679,13 +727,13 @@ impl State<ClientConnectionData> for ExpectCertificateVerify {
             })?;
 
         // 2. Verify their signature on the handshake.
-        let handshake_hash = self.transcript.get_current_hash();
+        let handshake_hash = self.transcript.current_hash();
         let sig_verified = self
             .config
             .verifier
             .verify_tls13_signature(
-                &verify::construct_tls13_server_verify_message(&handshake_hash),
-                &self.server_cert.cert_chain[0],
+                &construct_server_verify_message(&handshake_hash),
+                end_entity,
                 cert_verify,
             )
             .map_err(|err| {
@@ -693,7 +741,7 @@ impl State<ClientConnectionData> for ExpectCertificateVerify {
                     .send_cert_verify_error_alert(err)
             })?;
 
-        cx.common.peer_certificates = Some(self.server_cert.cert_chain);
+        cx.common.peer_certificates = Some(self.server_cert.cert_chain.into_owned());
         self.transcript.add_message(&m);
 
         Ok(Box::new(ExpectFinished {
@@ -708,6 +756,19 @@ impl State<ClientConnectionData> for ExpectCertificateVerify {
             sig_verified,
         }))
     }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        Box::new(ExpectCertificateVerify {
+            config: self.config,
+            server_name: self.server_name,
+            randoms: self.randoms,
+            suite: self.suite,
+            transcript: self.transcript,
+            key_schedule: self.key_schedule,
+            server_cert: self.server_cert.into_owned(),
+            client_auth: self.client_auth,
+        })
+    }
 }
 
 fn emit_certificate_tls13(
@@ -718,7 +779,7 @@ fn emit_certificate_tls13(
 ) {
     let context = auth_context.unwrap_or_default();
 
-    let mut cert_payload = CertificatePayloadTLS13 {
+    let mut cert_payload = CertificatePayloadTls13 {
         context: PayloadU8::new(context),
         entries: Vec::new(),
     };
@@ -735,7 +796,7 @@ fn emit_certificate_tls13(
         version: ProtocolVersion::TLSv1_3,
         payload: MessagePayload::handshake(HandshakeMessagePayload {
             typ: HandshakeType::Certificate,
-            payload: HandshakePayload::CertificateTLS13(cert_payload),
+            payload: HandshakePayload::CertificateTls13(cert_payload),
         }),
     };
     transcript.add_message(&m);
@@ -747,7 +808,7 @@ fn emit_certverify_tls13(
     signer: &dyn Signer,
     common: &mut CommonState,
 ) -> Result<(), Error> {
-    let message = verify::construct_tls13_client_verify_message(&transcript.get_current_hash());
+    let message = construct_client_verify_message(&transcript.current_hash());
 
     let scheme = signer.scheme();
     let sig = signer.sign(&message)?;
@@ -768,7 +829,7 @@ fn emit_certverify_tls13(
 
 fn emit_finished_tls13(
     transcript: &mut HandshakeHash,
-    verify_data: ring::hmac::Tag,
+    verify_data: &crypto::hmac::Tag,
     common: &mut CommonState,
 ) {
     let verify_data_payload = Payload::new(verify_data.as_ref());
@@ -804,7 +865,7 @@ fn emit_end_of_early_data_tls13(transcript: &mut HandshakeHash, common: &mut Com
 
 struct ExpectFinished {
     config: Arc<ClientConfig>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     randoms: ConnectionRandoms,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
@@ -815,27 +876,36 @@ struct ExpectFinished {
 }
 
 impl State<ClientConnectionData> for ExpectFinished {
-    fn handle(self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+    fn handle<'m>(
+        self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
         let mut st = *self;
         let finished =
             require_handshake_msg!(m, HandshakeType::Finished, HandshakePayload::Finished)?;
 
-        let handshake_hash = st.transcript.get_current_hash();
+        let handshake_hash = st.transcript.current_hash();
         let expect_verify_data = st
             .key_schedule
             .sign_server_finish(&handshake_hash);
 
-        let fin = constant_time::verify_slices_are_equal(expect_verify_data.as_ref(), &finished.0)
-            .map_err(|_| {
-                cx.common
-                    .send_fatal_alert(AlertDescription::DecryptError);
-                Error::DecryptError
-            })
-            .map(|_| verify::FinishedMessageVerified::assertion())?;
+        let fin = match ConstantTimeEq::ct_eq(expect_verify_data.as_ref(), finished.bytes()).into()
+        {
+            true => verify::FinishedMessageVerified::assertion(),
+            false => {
+                return Err(cx
+                    .common
+                    .send_fatal_alert(AlertDescription::DecryptError, Error::DecryptError));
+            }
+        };
 
         st.transcript.add_message(&m);
 
-        let hash_after_handshake = st.transcript.get_current_hash();
+        let hash_after_handshake = st.transcript.current_hash();
         /* The EndOfEarlyData message to server is still encrypted with early data keys,
          * but appears in the transcript after the server Finished. */
         if cx.common.early_traffic {
@@ -875,12 +945,12 @@ impl State<ClientConnectionData> for ExpectFinished {
             .key_schedule
             .into_pre_finished_client_traffic(
                 hash_after_handshake,
-                st.transcript.get_current_hash(),
+                st.transcript.current_hash(),
                 &*st.config.key_log,
                 &st.randoms.client,
             );
 
-        emit_finished_tls13(&mut st.transcript, verify_data, cx.common);
+        emit_finished_tls13(&mut st.transcript, &verify_data, cx.common);
 
         /* We're now sure this server supports TLS1.3.  But if we run out of TLS1.3 tickets
          * when connecting to it again, we definitely don't want to attempt a TLS1.2 resumption. */
@@ -892,9 +962,11 @@ impl State<ClientConnectionData> for ExpectFinished {
         /* Now move to our application traffic keys. */
         cx.common.check_aligned_handshake()?;
         let key_schedule_traffic = key_schedule_pre_finished.into_traffic(cx.common);
-        cx.common.start_traffic();
+        cx.common
+            .start_traffic(&mut cx.sendable_plaintext);
 
         let st = ExpectTraffic {
+            config: Arc::clone(&st.config),
             session_storage: Arc::clone(&st.config.resumption.store),
             server_name: st.server_name,
             suite: st.suite,
@@ -905,12 +977,14 @@ impl State<ClientConnectionData> for ExpectFinished {
             _fin_verified: fin,
         };
 
-        #[cfg(feature = "quic")]
-        if cx.common.is_quic() {
-            return Ok(Box::new(ExpectQuicTraffic(st)));
-        }
+        Ok(match cx.common.is_quic() {
+            true => Box::new(ExpectQuicTraffic(st)),
+            false => Box::new(st),
+        })
+    }
 
-        Ok(Box::new(st))
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
     }
 }
 
@@ -918,8 +992,9 @@ impl State<ClientConnectionData> for ExpectFinished {
 // In this state we can be sent tickets, key updates,
 // and application data.
 struct ExpectTraffic {
+    config: Arc<ClientConfig>,
     session_storage: Arc<dyn ClientSessionStore>,
-    server_name: ServerName,
+    server_name: ServerName<'static>,
     suite: &'static Tls13CipherSuite,
     transcript: HandshakeHash,
     key_schedule: KeyScheduleTraffic,
@@ -929,51 +1004,43 @@ struct ExpectTraffic {
 }
 
 impl ExpectTraffic {
-    #[allow(clippy::unnecessary_wraps)] // returns Err for #[cfg(feature = "quic")]
     fn handle_new_ticket_tls13(
         &mut self,
         cx: &mut ClientContext<'_>,
-        nst: &NewSessionTicketPayloadTLS13,
+        nst: &NewSessionTicketPayloadTls13,
     ) -> Result<(), Error> {
         if nst.has_duplicate_extension() {
-            cx.common
-                .send_fatal_alert(AlertDescription::IllegalParameter);
-            return Err(PeerMisbehaved::DuplicateNewSessionTicketExtensions.into());
+            return Err(cx.common.send_fatal_alert(
+                AlertDescription::IllegalParameter,
+                PeerMisbehaved::DuplicateNewSessionTicketExtensions,
+            ));
         }
 
-        let handshake_hash = self.transcript.get_current_hash();
+        let handshake_hash = self.transcript.current_hash();
         let secret = self
             .key_schedule
             .resumption_master_secret_and_derive_ticket_psk(&handshake_hash, &nst.nonce.0);
 
-        let time_now = match TimeBase::now() {
-            Ok(t) => t,
-            #[allow(unused_variables)]
-            Err(e) => {
-                debug!("Session not saved: {}", e);
-                return Ok(());
-            }
-        };
+        let now = self.config.current_time()?;
 
         #[allow(unused_mut)]
         let mut value = persist::Tls13ClientSessionValue::new(
             self.suite,
             nst.ticket.0.clone(),
-            secret,
+            secret.as_ref(),
             cx.common
                 .peer_certificates
                 .clone()
                 .unwrap_or_default(),
-            time_now,
+            now,
             nst.lifetime,
             nst.age_add,
-            nst.get_max_early_data_size()
+            nst.max_early_data_size()
                 .unwrap_or_default(),
         );
 
-        #[cfg(feature = "quic")]
         if cx.common.is_quic() {
-            if let Some(sz) = nst.get_max_early_data_size() {
+            if let Some(sz) = nst.max_early_data_size() {
                 if sz != 0 && sz != 0xffff_ffff {
                     return Err(PeerMisbehaved::InvalidMaxEarlyDataSize.into());
                 }
@@ -985,7 +1052,7 @@ impl ExpectTraffic {
         }
 
         self.session_storage
-            .insert_tls13_ticket(&self.server_name, value);
+            .insert_tls13_ticket(self.server_name.clone(), value);
         Ok(())
     }
 
@@ -994,13 +1061,11 @@ impl ExpectTraffic {
         common: &mut CommonState,
         key_update_request: &KeyUpdateRequest,
     ) -> Result<(), Error> {
-        #[cfg(feature = "quic")]
-        {
-            if let Protocol::Quic = common.protocol {
-                common.send_fatal_alert(AlertDescription::UnexpectedMessage);
-                warn!("KeyUpdate received in QUIC connection");
-                return Err(PeerMisbehaved::KeyUpdateReceivedInQuicConnection.into());
-            }
+        if let Protocol::Quic = common.protocol {
+            return Err(common.send_fatal_alert(
+                AlertDescription::UnexpectedMessage,
+                PeerMisbehaved::KeyUpdateReceivedInQuicConnection,
+            ));
         }
 
         // Mustn't be interleaved with other handshake messages.
@@ -1019,7 +1084,14 @@ impl ExpectTraffic {
 }
 
 impl State<ClientConnectionData> for ExpectTraffic {
-    fn handle(mut self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
         match m.payload {
             MessagePayload::ApplicationData(payload) => cx
                 .common
@@ -1027,7 +1099,7 @@ impl State<ClientConnectionData> for ExpectTraffic {
             MessagePayload::Handshake {
                 parsed:
                     HandshakeMessagePayload {
-                        payload: HandshakePayload::NewSessionTicketTLS13(ref new_ticket),
+                        payload: HandshakePayload::NewSessionTicketTls13(ref new_ticket),
                         ..
                     },
                 ..
@@ -1062,23 +1134,31 @@ impl State<ClientConnectionData> for ExpectTraffic {
             .export_keying_material(output, label, context)
     }
 
-    #[cfg(feature = "secret_extraction")]
     fn extract_secrets(&self) -> Result<PartiallyExtractedSecrets, Error> {
         self.key_schedule
             .extract_secrets(Side::Client)
     }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
 }
 
-#[cfg(feature = "quic")]
 struct ExpectQuicTraffic(ExpectTraffic);
 
-#[cfg(feature = "quic")]
 impl State<ClientConnectionData> for ExpectQuicTraffic {
-    fn handle(mut self: Box<Self>, cx: &mut ClientContext<'_>, m: Message) -> hs::NextStateOrError {
+    fn handle<'m>(
+        mut self: Box<Self>,
+        cx: &mut ClientContext<'_>,
+        m: Message<'m>,
+    ) -> hs::NextStateOrError<'m>
+    where
+        Self: 'm,
+    {
         let nst = require_handshake_msg!(
             m,
             HandshakeType::NewSessionTicket,
-            HandshakePayload::NewSessionTicketTLS13
+            HandshakePayload::NewSessionTicketTls13
         )?;
         self.0
             .handle_new_ticket_tls13(cx, nst)?;
@@ -1094,4 +1174,8 @@ impl State<ClientConnectionData> for ExpectQuicTraffic {
         self.0
             .export_keying_material(output, label, context)
     }
+
+    fn into_owned(self: Box<Self>) -> hs::NextState<'static> {
+        self
+    }
 }
