@@ -1,3 +1,8 @@
+use alloc::boxed::Box;
+use alloc::vec::Vec;
+
+use pki_types::CertificateDer;
+
 use crate::enums::{AlertDescription, ContentType, HandshakeType, ProtocolVersion};
 use crate::error::{Error, InvalidMessage, PeerMisbehaved};
 #[cfg(feature = "logging")]
@@ -7,20 +12,16 @@ use crate::msgs::base::Payload;
 use crate::msgs::enums::{AlertLevel, KeyUpdateRequest};
 use crate::msgs::fragmenter::MessageFragmenter;
 use crate::msgs::handshake::CertificateChain;
-use crate::msgs::message::MessagePayload;
-use crate::msgs::message::{BorrowedPlainMessage, Message, OpaqueMessage, PlainMessage};
-use crate::quic;
-use crate::record_layer;
-use crate::suites::PartiallyExtractedSecrets;
-use crate::suites::SupportedCipherSuite;
+use crate::msgs::message::{
+    Message, MessagePayload, OutboundChunks, OutboundOpaqueMessage, OutboundPlainMessage,
+    PlainMessage,
+};
+use crate::suites::{PartiallyExtractedSecrets, SupportedCipherSuite};
 #[cfg(feature = "tls12")]
 use crate::tls12::ConnectionSecrets;
+use crate::unbuffered::{EncryptError, InsufficientSizeError};
 use crate::vecbuf::ChunkVecBuffer;
-
-use alloc::boxed::Box;
-use alloc::vec::Vec;
-
-use pki_types::CertificateDer;
+use crate::{quic, record_layer};
 
 /// Connection state common to both client and server connections.
 pub struct CommonState {
@@ -36,12 +37,12 @@ pub struct CommonState {
     sent_fatal_alert: bool,
     /// If the peer has signaled end of stream.
     pub(crate) has_received_close_notify: bool,
+    #[cfg(feature = "std")]
     pub(crate) has_seen_eof: bool,
     pub(crate) received_middlebox_ccs: u8,
-    pub(crate) peer_certificates: Option<CertificateChain>,
+    pub(crate) peer_certificates: Option<CertificateChain<'static>>,
     message_fragmenter: MessageFragmenter,
     pub(crate) received_plaintext: ChunkVecBuffer,
-    sendable_plaintext: ChunkVecBuffer,
     pub(crate) sendable_tls: ChunkVecBuffer,
     queued_key_update_message: Option<Vec<u8>>,
 
@@ -65,12 +66,12 @@ impl CommonState {
             early_traffic: false,
             sent_fatal_alert: false,
             has_received_close_notify: false,
+            #[cfg(feature = "std")]
             has_seen_eof: false,
             received_middlebox_ccs: 0,
             peer_certificates: None,
             message_fragmenter: MessageFragmenter::default(),
             received_plaintext: ChunkVecBuffer::new(Some(DEFAULT_RECEIVED_PLAINTEXT_LIMIT)),
-            sendable_plaintext: ChunkVecBuffer::new(Some(DEFAULT_BUFFER_LIMIT)),
             sendable_tls: ChunkVecBuffer::new(Some(DEFAULT_BUFFER_LIMIT)),
             queued_key_update_message: None,
             protocol: Protocol::Tcp,
@@ -112,7 +113,7 @@ impl CommonState {
     /// if client authentication was completed.
     ///
     /// The return value is None until this value is available.
-    pub fn peer_certificates(&self) -> Option<&[CertificateDer<'_>]> {
+    pub fn peer_certificates(&self) -> Option<&[CertificateDer<'static>]> {
         self.peer_certificates.as_deref()
     }
 
@@ -148,6 +149,7 @@ impl CommonState {
         msg: Message,
         mut state: Box<dyn State<Data>>,
         data: &mut Data,
+        sendable_plaintext: Option<&mut ChunkVecBuffer>,
     ) -> Result<Box<dyn State<Data>>, Error> {
         // For TLS1.2, outside of the handshake, send rejection alerts for
         // renegotiation requests.  These can occur any time.
@@ -162,10 +164,14 @@ impl CommonState {
             }
         }
 
-        let mut cx = Context { common: self, data };
+        let mut cx = Context {
+            common: self,
+            data,
+            sendable_plaintext,
+        };
         match state.handle(&mut cx, msg) {
             Ok(next) => {
-                state = next;
+                state = next.into_owned();
                 Ok(state)
             }
             Err(e @ Error::InappropriateMessage { .. })
@@ -176,26 +182,51 @@ impl CommonState {
         }
     }
 
-    /// Send plaintext application data, fragmenting and
-    /// encrypting it as it goes out.
-    ///
-    /// If internal buffers are too small, this function will not accept
-    /// all the data.
-    pub(crate) fn send_some_plaintext(&mut self, data: &[u8]) -> usize {
-        self.perhaps_write_key_update();
-        self.send_plain(data, Limit::Yes)
-    }
-
-    pub(crate) fn send_early_plaintext(&mut self, data: &[u8]) -> usize {
-        debug_assert!(self.early_traffic);
-        debug_assert!(self.record_layer.is_encrypting());
-
-        if data.is_empty() {
-            // Don't send empty fragments.
-            return 0;
+    pub(crate) fn write_plaintext(
+        &mut self,
+        payload: OutboundChunks<'_>,
+        outgoing_tls: &mut [u8],
+    ) -> Result<usize, EncryptError> {
+        if payload.is_empty() {
+            return Ok(0);
         }
 
-        self.send_appdata_encrypt(data, Limit::Yes)
+        let fragments = self
+            .message_fragmenter
+            .fragment_payload(
+                ContentType::ApplicationData,
+                ProtocolVersion::TLSv1_2,
+                payload.clone(),
+            );
+
+        let remaining_encryptions = self
+            .record_layer
+            .remaining_write_seq()
+            .ok_or(EncryptError::EncryptExhausted)?;
+
+        if fragments.len() as u64 > remaining_encryptions.get() {
+            return Err(EncryptError::EncryptExhausted);
+        }
+
+        self.check_required_size(
+            outgoing_tls,
+            self.queued_key_update_message
+                .as_deref(),
+            fragments,
+        )?;
+
+        let fragments = self
+            .message_fragmenter
+            .fragment_payload(
+                ContentType::ApplicationData,
+                ProtocolVersion::TLSv1_2,
+                payload,
+            );
+
+        let opt_msg = self.queued_key_update_message.take();
+        let written = self.write_fragments(outgoing_tls, opt_msg, fragments);
+
+        Ok(written)
     }
 
     // Changing the keys must not span any fragmented handshake
@@ -225,23 +256,26 @@ impl CommonState {
     }
 
     /// Like send_msg_encrypt, but operate on an appdata directly.
-    fn send_appdata_encrypt(&mut self, payload: &[u8], limit: Limit) -> usize {
+    fn send_appdata_encrypt(&mut self, payload: OutboundChunks<'_>, limit: Limit) -> usize {
         // Here, the limit on sendable_tls applies to encrypted data,
         // but we're respecting it for plaintext data -- so we'll
         // be out by whatever the cipher+record overhead is.  That's a
         // constant and predictable amount, so it's not a terrible issue.
         let len = match limit {
+            #[cfg(feature = "std")]
             Limit::Yes => self
                 .sendable_tls
                 .apply_limit(payload.len()),
             Limit::No => payload.len(),
         };
 
-        let iter = self.message_fragmenter.fragment_slice(
-            ContentType::ApplicationData,
-            ProtocolVersion::TLSv1_2,
-            &payload[..len],
-        );
+        let iter = self
+            .message_fragmenter
+            .fragment_payload(
+                ContentType::ApplicationData,
+                ProtocolVersion::TLSv1_2,
+                payload.split_at(len).0,
+            );
         for m in iter {
             self.send_single_fragment(m);
         }
@@ -249,7 +283,7 @@ impl CommonState {
         len
     }
 
-    fn send_single_fragment(&mut self, m: BorrowedPlainMessage) {
+    fn send_single_fragment(&mut self, m: OutboundPlainMessage) {
         // Close connection once we start to run out of
         // sequence space.
         if self
@@ -269,108 +303,53 @@ impl CommonState {
         self.queue_tls_message(em);
     }
 
-    /// Encrypt and send some plaintext `data`.  `limit` controls
-    /// whether the per-connection buffer limits apply.
-    ///
-    /// Returns the number of bytes written from `data`: this might
-    /// be less than `data.len()` if buffer limits were exceeded.
-    fn send_plain(&mut self, data: &[u8], limit: Limit) -> usize {
-        if !self.may_send_application_data {
-            // If we haven't completed handshaking, buffer
-            // plaintext to send once we do.
-            let len = match limit {
-                Limit::Yes => self
-                    .sendable_plaintext
-                    .append_limited_copy(data),
-                Limit::No => self
-                    .sendable_plaintext
-                    .append(data.to_vec()),
-            };
-            return len;
-        }
-
+    fn send_plain_non_buffering(&mut self, payload: OutboundChunks<'_>, limit: Limit) -> usize {
+        debug_assert!(self.may_send_application_data);
         debug_assert!(self.record_layer.is_encrypting());
 
-        if data.is_empty() {
+        if payload.is_empty() {
             // Don't send empty fragments.
             return 0;
         }
 
-        self.send_appdata_encrypt(data, limit)
+        self.send_appdata_encrypt(payload, limit)
     }
 
-    pub(crate) fn start_outgoing_traffic(&mut self) {
+    /// Mark the connection as ready to send application data.
+    ///
+    /// Also flush `sendable_plaintext` if it is `Some`.  
+    pub(crate) fn start_outgoing_traffic(
+        &mut self,
+        sendable_plaintext: &mut Option<&mut ChunkVecBuffer>,
+    ) {
         self.may_send_application_data = true;
-        self.flush_plaintext();
+        if let Some(sendable_plaintext) = sendable_plaintext {
+            self.flush_plaintext(sendable_plaintext);
+        }
     }
 
-    pub(crate) fn start_traffic(&mut self) {
+    /// Mark the connection as ready to send and receive application data.
+    ///
+    /// Also flush `sendable_plaintext` if it is `Some`.  
+    pub(crate) fn start_traffic(&mut self, sendable_plaintext: &mut Option<&mut ChunkVecBuffer>) {
         self.may_receive_application_data = true;
-        self.start_outgoing_traffic();
-    }
-
-    /// Sets a limit on the internal buffers used to buffer
-    /// unsent plaintext (prior to completing the TLS handshake)
-    /// and unsent TLS records.  This limit acts only on application
-    /// data written through [`Connection::writer`].
-    ///
-    /// By default the limit is 64KB.  The limit can be set
-    /// at any time, even if the current buffer use is higher.
-    ///
-    /// [`None`] means no limit applies, and will mean that written
-    /// data is buffered without bound -- it is up to the application
-    /// to appropriately schedule its plaintext and TLS writes to bound
-    /// memory usage.
-    ///
-    /// For illustration: `Some(1)` means a limit of one byte applies:
-    /// [`Connection::writer`] will accept only one byte, encrypt it and
-    /// add a TLS header.  Once this is sent via [`Connection::write_tls`],
-    /// another byte may be sent.
-    ///
-    /// # Internal write-direction buffering
-    /// rustls has two buffers whose size are bounded by this setting:
-    ///
-    /// ## Buffering of unsent plaintext data prior to handshake completion
-    ///
-    /// Calls to [`Connection::writer`] before or during the handshake
-    /// are buffered (up to the limit specified here).  Once the
-    /// handshake completes this data is encrypted and the resulting
-    /// TLS records are added to the outgoing buffer.
-    ///
-    /// ## Buffering of outgoing TLS records
-    ///
-    /// This buffer is used to store TLS records that rustls needs to
-    /// send to the peer.  It is used in these two circumstances:
-    ///
-    /// - by [`Connection::process_new_packets`] when a handshake or alert
-    ///   TLS record needs to be sent.
-    /// - by [`Connection::writer`] post-handshake: the plaintext is
-    ///   encrypted and the resulting TLS record is buffered.
-    ///
-    /// This buffer is emptied by [`Connection::write_tls`].
-    ///
-    /// [`Connection::writer`]: crate::Connection::writer
-    /// [`Connection::write_tls`]: crate::Connection::write_tls
-    /// [`Connection::process_new_packets`]: crate::Connection::process_new_packets
-    pub fn set_buffer_limit(&mut self, limit: Option<usize>) {
-        self.sendable_plaintext.set_limit(limit);
-        self.sendable_tls.set_limit(limit);
+        self.start_outgoing_traffic(sendable_plaintext);
     }
 
     /// Send any buffered plaintext.  Plaintext is buffered if
     /// written during handshake.
-    fn flush_plaintext(&mut self) {
+    fn flush_plaintext(&mut self, sendable_plaintext: &mut ChunkVecBuffer) {
         if !self.may_send_application_data {
             return;
         }
 
-        while let Some(buf) = self.sendable_plaintext.pop() {
-            self.send_plain(&buf, Limit::No);
+        while let Some(buf) = sendable_plaintext.pop() {
+            self.send_plain_non_buffering(buf.as_slice().into(), Limit::No);
         }
     }
 
     // Put m into sendable_tls for writing.
-    fn queue_tls_message(&mut self, m: OpaqueMessage) {
+    fn queue_tls_message(&mut self, m: OutboundOpaqueMessage) {
         self.sendable_tls.append(m.encode());
     }
 
@@ -408,7 +387,8 @@ impl CommonState {
     }
 
     pub(crate) fn take_received_plaintext(&mut self, bytes: Payload) {
-        self.received_plaintext.append(bytes.0);
+        self.received_plaintext
+            .append(bytes.into_vec());
     }
 
     #[cfg(feature = "tls12")]
@@ -452,7 +432,7 @@ impl CommonState {
             if self.is_tls13() && alert.description != AlertDescription::UserCanceled {
                 return Err(self.send_fatal_alert(AlertDescription::DecodeError, err));
             } else {
-                warn!("TLS alert warning received: {:#?}", alert);
+                warn!("TLS alert warning received: {:?}", alert);
                 return Ok(());
             }
         }
@@ -493,11 +473,88 @@ impl CommonState {
         self.send_warning_alert_no_log(AlertDescription::CloseNotify);
     }
 
+    pub(crate) fn eager_send_close_notify(
+        &mut self,
+        outgoing_tls: &mut [u8],
+    ) -> Result<usize, EncryptError> {
+        debug_assert!(self.record_layer.is_encrypting());
+
+        let m = Message::build_alert(AlertLevel::Warning, AlertDescription::CloseNotify).into();
+
+        let iter = self
+            .message_fragmenter
+            .fragment_message(&m);
+
+        self.check_required_size(outgoing_tls, None, iter)?;
+
+        debug!("Sending warning alert {:?}", AlertDescription::CloseNotify);
+
+        let iter = self
+            .message_fragmenter
+            .fragment_message(&m);
+
+        let written = self.write_fragments(outgoing_tls, None, iter);
+
+        Ok(written)
+    }
+
     fn send_warning_alert_no_log(&mut self, desc: AlertDescription) {
         let m = Message::build_alert(AlertLevel::Warning, desc);
         self.send_msg(m, self.record_layer.is_encrypting());
     }
 
+    fn check_required_size<'a>(
+        &self,
+        outgoing_tls: &mut [u8],
+        opt_msg: Option<&[u8]>,
+        fragments: impl Iterator<Item = OutboundPlainMessage<'a>>,
+    ) -> Result<(), EncryptError> {
+        let mut required_size = 0;
+        if let Some(message) = opt_msg {
+            required_size += message.len();
+        }
+
+        for m in fragments {
+            required_size += m.encoded_len(&self.record_layer);
+        }
+
+        if required_size > outgoing_tls.len() {
+            return Err(EncryptError::InsufficientSize(InsufficientSizeError {
+                required_size,
+            }));
+        }
+
+        Ok(())
+    }
+
+    fn write_fragments<'a>(
+        &mut self,
+        outgoing_tls: &mut [u8],
+        opt_msg: Option<Vec<u8>>,
+        fragments: impl Iterator<Item = OutboundPlainMessage<'a>>,
+    ) -> usize {
+        let mut written = 0;
+
+        if let Some(message) = opt_msg {
+            let len = message.len();
+            outgoing_tls[written..written + len].copy_from_slice(&message);
+            written += len;
+        }
+
+        for m in fragments {
+            let em = self
+                .record_layer
+                .encrypt_outgoing(m)
+                .encode();
+
+            let len = em.len();
+            outgoing_tls[written..written + len].copy_from_slice(&em);
+            written += len;
+        }
+
+        written
+    }
+
     pub(crate) fn set_max_fragment_size(&mut self, new: Option<usize>) -> Result<(), Error> {
         self.message_fragmenter
             .set_max_fragment_size(new)
@@ -560,10 +617,63 @@ impl CommonState {
         let message = PlainMessage::from(Message::build_key_update_notify());
         self.queued_key_update_message = Some(
             self.record_layer
-                .encrypt_outgoing(message.borrow())
+                .encrypt_outgoing(message.borrow_outbound())
                 .encode(),
         );
     }
+}
+
+#[cfg(feature = "std")]
+impl CommonState {
+    /// Send plaintext application data, fragmenting and
+    /// encrypting it as it goes out.
+    ///
+    /// If internal buffers are too small, this function will not accept
+    /// all the data.
+    pub(crate) fn buffer_plaintext(
+        &mut self,
+        payload: OutboundChunks<'_>,
+        sendable_plaintext: &mut ChunkVecBuffer,
+    ) -> usize {
+        self.perhaps_write_key_update();
+        self.send_plain(payload, Limit::Yes, sendable_plaintext)
+    }
+
+    pub(crate) fn send_early_plaintext(&mut self, data: &[u8]) -> usize {
+        debug_assert!(self.early_traffic);
+        debug_assert!(self.record_layer.is_encrypting());
+
+        if data.is_empty() {
+            // Don't send empty fragments.
+            return 0;
+        }
+
+        self.send_appdata_encrypt(data.into(), Limit::Yes)
+    }
+
+    /// Encrypt and send some plaintext `data`.  `limit` controls
+    /// whether the per-connection buffer limits apply.
+    ///
+    /// Returns the number of bytes written from `data`: this might
+    /// be less than `data.len()` if buffer limits were exceeded.
+    fn send_plain(
+        &mut self,
+        payload: OutboundChunks<'_>,
+        limit: Limit,
+        sendable_plaintext: &mut ChunkVecBuffer,
+    ) -> usize {
+        if !self.may_send_application_data {
+            // If we haven't completed handshaking, buffer
+            // plaintext to send once we do.
+            let len = match limit {
+                Limit::Yes => sendable_plaintext.append_limited_copy(payload),
+                Limit::No => sendable_plaintext.append(payload.to_vec()),
+            };
+            return len;
+        }
+
+        self.send_plain_non_buffering(payload, limit)
+    }
 
     pub(crate) fn perhaps_write_key_update(&mut self) {
         if let Some(message) = self.queued_key_update_message.take() {
@@ -612,11 +722,13 @@ impl IoState {
 }
 
 pub(crate) trait State<Data>: Send + Sync {
-    fn handle(
+    fn handle<'m>(
         self: Box<Self>,
         cx: &mut Context<'_, Data>,
-        message: Message,
-    ) -> Result<Box<dyn State<Data>>, Error>;
+        message: Message<'m>,
+    ) -> Result<Box<dyn State<Data> + 'm>, Error>
+    where
+        Self: 'm;
 
     fn export_keying_material(
         &self,
@@ -630,11 +742,18 @@ pub(crate) trait State<Data>: Send + Sync {
     fn extract_secrets(&self) -> Result<PartiallyExtractedSecrets, Error> {
         Err(Error::HandshakeNotComplete)
     }
+
+    fn handle_decrypt_error(&self) {}
+
+    fn into_owned(self: Box<Self>) -> Box<dyn State<Data> + 'static>;
 }
 
 pub(crate) struct Context<'a, Data> {
     pub(crate) common: &'a mut CommonState,
     pub(crate) data: &'a mut Data,
+    /// Buffered plaintext. This is `Some` if any plaintext was written during handshake and `None`
+    /// otherwise.
+    pub(crate) sendable_plaintext: Option<&'a mut ChunkVecBuffer>,
 }
 
 /// Side of the connection.
@@ -662,9 +781,10 @@ pub(crate) enum Protocol {
 }
 
 enum Limit {
+    #[cfg(feature = "std")]
     Yes,
     No,
 }
 
 const DEFAULT_RECEIVED_PLAINTEXT_LIMIT: usize = 16 * 1024;
-const DEFAULT_BUFFER_LIMIT: usize = 64 * 1024;
+pub(crate) const DEFAULT_BUFFER_LIMIT: usize = 64 * 1024;
