@@ -1,22 +1,21 @@
 #![allow(clippy::duplicate_mod)]
 
-use crate::enums::{SignatureAlgorithm, SignatureScheme};
-use crate::error::Error;
-use crate::sign::{Signer, SigningKey};
-use crate::x509::{asn1_wrap, wrap_in_sequence};
+use alloc::boxed::Box;
+use alloc::string::ToString;
+use alloc::sync::Arc;
+use alloc::vec::Vec;
+use alloc::{format, vec};
+use core::fmt::{self, Debug, Formatter};
+
+use pki_types::{PrivateKeyDer, PrivatePkcs8KeyDer};
 
 use super::ring_like::io::der;
 use super::ring_like::rand::{SecureRandom, SystemRandom};
 use super::ring_like::signature::{self, EcdsaKeyPair, Ed25519KeyPair, RsaKeyPair};
-use pki_types::{PrivateKeyDer, PrivatePkcs8KeyDer};
-
-use alloc::boxed::Box;
-use alloc::format;
-use alloc::string::ToString;
-use alloc::sync::Arc;
-use alloc::vec;
-use alloc::vec::Vec;
-use core::fmt::{self, Debug, Formatter};
+use crate::enums::{SignatureAlgorithm, SignatureScheme};
+use crate::error::Error;
+use crate::sign::{Signer, SigningKey};
+use crate::x509::{asn1_wrap, wrap_in_sequence};
 
 /// Parse `der` as any supported key encoding/type, returning
 /// the first which works.
@@ -165,7 +164,7 @@ impl RsaSigner {
 
 impl Signer for RsaSigner {
     fn sign(&self, message: &[u8]) -> Result<Vec<u8>, Error> {
-        let mut sig = vec![0; super::ring_shim::rsa_key_pair_public_modulus_len(&self.key)];
+        let mut sig = vec![0; self.key.public().modulus_len()];
 
         let rng = SystemRandom::new();
         self.key
@@ -218,7 +217,7 @@ impl EcdsaSigningKey {
                 Self::convert_sec1_to_pkcs8(scheme, sigalg, sec1.secret_sec1_der(), &rng)?
             }
             PrivateKeyDer::Pkcs8(pkcs8) => {
-                super::ring_shim::ecdsa_key_pair_from_pkcs8(sigalg, pkcs8.secret_pkcs8_der(), &rng)?
+                EcdsaKeyPair::from_pkcs8(sigalg, pkcs8.secret_pkcs8_der(), &rng).map_err(|_| ())?
             }
             _ => return Err(()),
         };
@@ -250,7 +249,7 @@ impl EcdsaSigningKey {
         pkcs8_inner.extend_from_slice(pkcs8_prefix);
         pkcs8_inner.extend_from_slice(&sec1_wrap);
 
-        super::ring_shim::ecdsa_key_pair_from_pkcs8(sigalg, &wrap_in_sequence(&pkcs8_inner), rng)
+        EcdsaKeyPair::from_pkcs8(sigalg, &wrap_in_sequence(&pkcs8_inner), rng).map_err(|_| ())
     }
 }
 
@@ -306,7 +305,7 @@ struct EcdsaSigner {
 
 impl Signer for EcdsaSigner {
     fn sign(&self, message: &[u8]) -> Result<Vec<u8>, Error> {
-        let rng = super::ring_like::rand::SystemRandom::new();
+        let rng = SystemRandom::new();
         self.key
             .sign(&rng, message)
             .map_err(|_| Error::General("signing failed".into()))
@@ -408,9 +407,12 @@ impl Debug for Ed25519Signer {
 
 #[cfg(test)]
 mod tests {
-    use super::*;
+    use alloc::format;
+
     use pki_types::{PrivatePkcs1KeyDer, PrivateSec1KeyDer};
 
+    use super::*;
+
     #[test]
     fn can_load_ecdsa_nistp256_pkcs8() {
         let key =
@@ -430,6 +432,37 @@ mod tests {
         assert!(any_ecdsa_type(&key).is_ok());
     }
 
+    #[test]
+    fn can_sign_ecdsa_nistp256() {
+        let key = PrivateKeyDer::Sec1(PrivateSec1KeyDer::from(
+            &include_bytes!("../../testdata/nistp256key.der")[..],
+        ));
+
+        let k = any_supported_type(&key).unwrap();
+        assert_eq!(format!("{:?}", k), "EcdsaSigningKey { algorithm: ECDSA }");
+        assert_eq!(k.algorithm(), SignatureAlgorithm::ECDSA);
+
+        assert!(k
+            .choose_scheme(&[SignatureScheme::RSA_PKCS1_SHA256])
+            .is_none());
+        assert!(k
+            .choose_scheme(&[SignatureScheme::ECDSA_NISTP384_SHA384])
+            .is_none());
+        let s = k
+            .choose_scheme(&[SignatureScheme::ECDSA_NISTP256_SHA256])
+            .unwrap();
+        assert_eq!(
+            format!("{:?}", s),
+            "EcdsaSigner { scheme: ECDSA_NISTP256_SHA256 }"
+        );
+        assert_eq!(s.scheme(), SignatureScheme::ECDSA_NISTP256_SHA256);
+        // nb. signature is variable length and asn.1-encoded
+        assert!(s
+            .sign(b"hello")
+            .unwrap()
+            .starts_with(&[0x30]));
+    }
+
     #[test]
     fn can_load_ecdsa_nistp384_pkcs8() {
         let key =
@@ -449,6 +482,37 @@ mod tests {
         assert!(any_ecdsa_type(&key).is_ok());
     }
 
+    #[test]
+    fn can_sign_ecdsa_nistp384() {
+        let key = PrivateKeyDer::Sec1(PrivateSec1KeyDer::from(
+            &include_bytes!("../../testdata/nistp384key.der")[..],
+        ));
+
+        let k = any_supported_type(&key).unwrap();
+        assert_eq!(format!("{:?}", k), "EcdsaSigningKey { algorithm: ECDSA }");
+        assert_eq!(k.algorithm(), SignatureAlgorithm::ECDSA);
+
+        assert!(k
+            .choose_scheme(&[SignatureScheme::RSA_PKCS1_SHA256])
+            .is_none());
+        assert!(k
+            .choose_scheme(&[SignatureScheme::ECDSA_NISTP256_SHA256])
+            .is_none());
+        let s = k
+            .choose_scheme(&[SignatureScheme::ECDSA_NISTP384_SHA384])
+            .unwrap();
+        assert_eq!(
+            format!("{:?}", s),
+            "EcdsaSigner { scheme: ECDSA_NISTP384_SHA384 }"
+        );
+        assert_eq!(s.scheme(), SignatureScheme::ECDSA_NISTP384_SHA384);
+        // nb. signature is variable length and asn.1-encoded
+        assert!(s
+            .sign(b"hello")
+            .unwrap()
+            .starts_with(&[0x30]));
+    }
+
     #[test]
     fn can_load_eddsa_pkcs8() {
         let key = PrivatePkcs8KeyDer::from(&include_bytes!("../../testdata/eddsakey.der")[..]);
@@ -458,6 +522,31 @@ mod tests {
         assert!(any_ecdsa_type(&key).is_err());
     }
 
+    #[test]
+    fn can_sign_eddsa() {
+        let key = PrivatePkcs8KeyDer::from(&include_bytes!("../../testdata/eddsakey.der")[..]);
+
+        let k = any_eddsa_type(&key).unwrap();
+        assert_eq!(
+            format!("{:?}", k),
+            "Ed25519SigningKey { algorithm: ED25519 }"
+        );
+        assert_eq!(k.algorithm(), SignatureAlgorithm::ED25519);
+
+        assert!(k
+            .choose_scheme(&[SignatureScheme::RSA_PKCS1_SHA256])
+            .is_none());
+        assert!(k
+            .choose_scheme(&[SignatureScheme::ECDSA_NISTP256_SHA256])
+            .is_none());
+        let s = k
+            .choose_scheme(&[SignatureScheme::ED25519])
+            .unwrap();
+        assert_eq!(format!("{:?}", s), "Ed25519Signer { scheme: ED25519 }");
+        assert_eq!(s.scheme(), SignatureScheme::ED25519);
+        assert_eq!(s.sign(b"hello").unwrap().len(), 64);
+    }
+
     #[test]
     fn can_load_rsa2048_pkcs8() {
         let key =
@@ -477,6 +566,42 @@ mod tests {
         assert!(any_ecdsa_type(&key).is_err());
     }
 
+    #[test]
+    fn can_sign_rsa2048() {
+        let key = PrivateKeyDer::Pkcs8(PrivatePkcs8KeyDer::from(
+            &include_bytes!("../../testdata/rsa2048key.pkcs8.der")[..],
+        ));
+
+        let k = any_supported_type(&key).unwrap();
+        assert_eq!(format!("{:?}", k), "RsaSigningKey { algorithm: RSA }");
+        assert_eq!(k.algorithm(), SignatureAlgorithm::RSA);
+
+        assert!(k
+            .choose_scheme(&[SignatureScheme::ECDSA_NISTP256_SHA256])
+            .is_none());
+        assert!(k
+            .choose_scheme(&[SignatureScheme::ED25519])
+            .is_none());
+
+        let s = k
+            .choose_scheme(&[SignatureScheme::RSA_PSS_SHA256])
+            .unwrap();
+        assert_eq!(format!("{:?}", s), "RsaSigner { scheme: RSA_PSS_SHA256 }");
+        assert_eq!(s.scheme(), SignatureScheme::RSA_PSS_SHA256);
+        assert_eq!(s.sign(b"hello").unwrap().len(), 256);
+
+        for scheme in &[
+            SignatureScheme::RSA_PKCS1_SHA256,
+            SignatureScheme::RSA_PKCS1_SHA384,
+            SignatureScheme::RSA_PKCS1_SHA512,
+            SignatureScheme::RSA_PSS_SHA256,
+            SignatureScheme::RSA_PSS_SHA384,
+            SignatureScheme::RSA_PSS_SHA512,
+        ] {
+            k.choose_scheme(&[*scheme]).unwrap();
+        }
+    }
+
     #[test]
     fn cannot_load_invalid_pkcs8_encoding() {
         let key = PrivateKeyDer::Pkcs8(PrivatePkcs8KeyDer::from(&b"invalid"[..]));
