@@ -5,13 +5,19 @@ use pki_types::{CertificateDer, CertificateRevocationListDer, UnixTime};
 use webpki::{CertRevocationList, RevocationCheckDepth, UnknownStatusPolicy};
 
 use super::{pki_error, VerifierBuilderError};
+#[cfg(doc)]
+use crate::crypto;
 use crate::crypto::{CryptoProvider, WebPkiSupportedAlgorithms};
+#[cfg(doc)]
+use crate::server::ServerConfig;
 use crate::verify::{
     ClientCertVerified, ClientCertVerifier, DigitallySignedStruct, HandshakeSignatureValid,
     NoClientAuth,
 };
 use crate::webpki::parse_crls;
 use crate::webpki::verify::{verify_tls12_signature, verify_tls13_signature, ParsedCertificate};
+#[cfg(doc)]
+use crate::ConfigBuilder;
 use crate::{DistinguishedName, Error, RootCertStore, SignatureScheme};
 
 /// A builder for configuring a `webpki` client certificate verifier.
@@ -138,11 +144,11 @@ impl ClientCertVerifierBuilder {
     /// certificate authentication offer with a client certificate.
     ///
     /// If `with_signature_verification_algorithms` was not called on the builder, a default set of
-    /// signature verification algorithms is used, controlled by the selected [`crate::crypto::CryptoProvider`].
+    /// signature verification algorithms is used, controlled by the selected [`CryptoProvider`].
     ///
     /// Once built, the provided `Arc<dyn ClientCertVerifier>` can be used with a Rustls
-    /// [crate::server::ServerConfig] to configure client certificate validation using
-    /// [`with_client_cert_verifier`][crate::ConfigBuilder<ClientConfig, WantsVerifier>::with_client_cert_verifier].
+    /// [`ServerConfig`] to configure client certificate validation using
+    /// [`with_client_cert_verifier`][ConfigBuilder<ClientConfig, WantsVerifier>::with_client_cert_verifier].
     ///
     /// # Errors
     /// This function will return a `ClientCertVerifierBuilderError` if:
@@ -166,16 +172,17 @@ impl ClientCertVerifierBuilder {
 }
 
 /// A client certificate verifier that uses the `webpki` crate[^1] to perform client certificate
-/// validation. It must be created via the [WebPkiClientVerifier::builder()] function.
+/// validation. It must be created via the [`WebPkiClientVerifier::builder()`] or
+/// [`WebPkiClientVerifier::builder_with_provider()`] functions.
 ///
-/// Once built, the provided `Arc<dyn ClientCertVerifier>` can be used with a Rustls [crate::server::ServerConfig]
-/// to configure client certificate validation using [`with_client_cert_verifier`][crate::ConfigBuilder<ClientConfig, WantsVerifier>::with_client_cert_verifier].
+/// Once built, the provided `Arc<dyn ClientCertVerifier>` can be used with a Rustls [`ServerConfig`]
+/// to configure client certificate validation using [`with_client_cert_verifier`][ConfigBuilder<ClientConfig, WantsVerifier>::with_client_cert_verifier].
 ///
 /// Example:
 ///
 /// To require all clients present a client certificate issued by a trusted CA:
 /// ```no_run
-/// # #[cfg(feature = "ring")] {
+/// # #[cfg(any(feature = "ring", feature = "aws_lc_rs"))] {
 /// # use rustls::RootCertStore;
 /// # use rustls::server::WebPkiClientVerifier;
 /// # let roots = RootCertStore::empty();
@@ -188,7 +195,7 @@ impl ClientCertVerifierBuilder {
 /// Or, to allow clients presenting a client certificate authenticated by a trusted CA, or
 /// anonymous clients that present no client certificate:
 /// ```no_run
-/// # #[cfg(feature = "ring")] {
+/// # #[cfg(any(feature = "ring", feature = "aws_lc_rs"))] {
 /// # use rustls::RootCertStore;
 /// # use rustls::server::WebPkiClientVerifier;
 /// # let roots = RootCertStore::empty();
@@ -210,7 +217,7 @@ impl ClientCertVerifierBuilder {
 /// You can also configure the client verifier to check for certificate revocation with
 /// client certificate revocation lists (CRLs):
 /// ```no_run
-/// # #[cfg(feature = "ring")] {
+/// # #[cfg(any(feature = "ring", feature = "aws_lc_rs"))] {
 /// # use rustls::RootCertStore;
 /// # use rustls::server::{WebPkiClientVerifier};
 /// # let roots = RootCertStore::empty();
@@ -236,19 +243,20 @@ pub struct WebPkiClientVerifier {
 
 impl WebPkiClientVerifier {
     /// Create a builder for the `webpki` client certificate verifier configuration using
-    /// the default [`CryptoProvider`].
+    /// the [process-default `CryptoProvider`][CryptoProvider#using-the-per-process-default-cryptoprovider].
     ///
     /// Client certificate authentication will be offered by the server, and client certificates
     /// will be verified using the trust anchors found in the provided `roots`. If you
-    /// wish to disable client authentication use [WebPkiClientVerifier::no_client_auth()] instead.
+    /// wish to disable client authentication use [`WebPkiClientVerifier::no_client_auth()`] instead.
     ///
-    /// The cryptography used comes from the default [`CryptoProvider`]: [`crate::crypto::ring::default_provider`].
-    /// Use [`Self::builder_with_provider`] if you wish to customize this.
+    /// Use [`Self::builder_with_provider`] if you wish to specify an explicit provider.
     ///
     /// For more information, see the [`ClientCertVerifierBuilder`] documentation.
-    #[cfg(feature = "ring")]
     pub fn builder(roots: Arc<RootCertStore>) -> ClientCertVerifierBuilder {
-        Self::builder_with_provider(roots, crate::crypto::ring::default_provider().into())
+        Self::builder_with_provider(
+            roots,
+            Arc::clone(CryptoProvider::get_default_or_install_from_crate_features()),
+        )
     }
 
     /// Create a builder for the `webpki` client certificate verifier configuration using
@@ -398,15 +406,16 @@ pub(crate) enum AnonymousClientPolicy {
     Deny,
 }
 
-#[cfg(all(test, feature = "ring"))]
-mod tests {
+test_for_each_provider! {
     use super::WebPkiClientVerifier;
     use crate::server::VerifierBuilderError;
     use crate::RootCertStore;
 
     use pki_types::{CertificateDer, CertificateRevocationListDer};
 
+    use std::prelude::v1::*;
     use std::sync::Arc;
+    use std::{vec, format, println};
 
     fn load_crls(crls_der: &[&[u8]]) -> Vec<CertificateRevocationListDer<'static>> {
         crls_der
@@ -422,7 +431,7 @@ mod tests {
 
     fn test_crls() -> Vec<CertificateRevocationListDer<'static>> {
         load_crls(&[
-            include_bytes!("../../../test-ca/ecdsa/client.revoked.crl.pem").as_slice(),
+            include_bytes!("../../../test-ca/ecdsa-p256/client.revoked.crl.pem").as_slice(),
             include_bytes!("../../../test-ca/rsa/client.revoked.crl.pem").as_slice(),
         ])
     }
@@ -439,7 +448,7 @@ mod tests {
 
     fn test_roots() -> Arc<RootCertStore> {
         load_roots(&[
-            include_bytes!("../../../test-ca/ecdsa/ca.der").as_slice(),
+            include_bytes!("../../../test-ca/ecdsa-p256/ca.der").as_slice(),
             include_bytes!("../../../test-ca/rsa/ca.der").as_slice(),
         ])
     }
@@ -454,7 +463,10 @@ mod tests {
     fn test_client_verifier_required_auth() {
         // We should be able to build a verifier that requires client authentication, and does
         // no revocation checking.
-        let builder = WebPkiClientVerifier::builder(test_roots());
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        );
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -464,7 +476,11 @@ mod tests {
     fn test_client_verifier_optional_auth() {
         // We should be able to build a verifier that allows client authentication, and anonymous
         // access, and does no revocation checking.
-        let builder = WebPkiClientVerifier::builder(test_roots()).allow_unauthenticated();
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .allow_unauthenticated();
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -475,7 +491,10 @@ mod tests {
         // We should be able to build a verifier that requires client authentication, and does
         // no revocation checking, that hasn't been configured to determine how to handle
         // unauthenticated clients yet.
-        let builder = WebPkiClientVerifier::builder(test_roots());
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        );
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -485,7 +504,11 @@ mod tests {
     fn test_client_verifier_without_crls_opptional_auth() {
         // We should be able to build a verifier that allows client authentication,
         // and anonymous access, that does no revocation checking.
-        let builder = WebPkiClientVerifier::builder(test_roots()).allow_unauthenticated();
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .allow_unauthenticated();
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -494,9 +517,12 @@ mod tests {
     #[test]
     fn test_with_invalid_crls() {
         // Trying to build a client verifier with invalid CRLs should error at build time.
-        let result = WebPkiClientVerifier::builder(test_roots())
-            .with_crls(vec![CertificateRevocationListDer::from(vec![0xFF])])
-            .build();
+        let result = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .with_crls(vec![CertificateRevocationListDer::from(vec![0xFF])])
+        .build();
         assert!(matches!(result, Err(VerifierBuilderError::InvalidCrl(_))));
     }
 
@@ -508,9 +534,12 @@ mod tests {
             load_crls(&[
                 include_bytes!("../../../test-ca/eddsa/client.revoked.crl.pem").as_slice(),
             ]);
-        let builder = WebPkiClientVerifier::builder(test_roots())
-            .with_crls(initial_crls.clone())
-            .with_crls(extra_crls.clone());
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .with_crls(initial_crls.clone())
+        .with_crls(extra_crls.clone());
 
         // There should be the expected number of crls.
         assert_eq!(builder.crls.len(), initial_crls.len() + extra_crls.len());
@@ -523,7 +552,11 @@ mod tests {
     fn test_client_verifier_with_crls_required_auth_implicit() {
         // We should be able to build a verifier that requires client authentication, and that does
         // revocation checking with CRLs, and that does not allow any anonymous access.
-        let builder = WebPkiClientVerifier::builder(test_roots()).with_crls(test_crls());
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .with_crls(test_crls());
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -533,9 +566,12 @@ mod tests {
     fn test_client_verifier_with_crls_optional_auth() {
         // We should be able to build a verifier that supports client authentication, that does
         // revocation checking with CRLs, and that allows anonymous access.
-        let builder = WebPkiClientVerifier::builder(test_roots())
-            .with_crls(test_crls())
-            .allow_unauthenticated();
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .with_crls(test_crls())
+        .allow_unauthenticated();
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -544,9 +580,12 @@ mod tests {
     #[test]
     fn test_client_verifier_ee_only() {
         // We should be able to build a client verifier that only checks EE revocation status.
-        let builder = WebPkiClientVerifier::builder(test_roots())
-            .with_crls(test_crls())
-            .only_check_end_entity_revocation();
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .with_crls(test_crls())
+        .only_check_end_entity_revocation();
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -555,9 +594,12 @@ mod tests {
     #[test]
     fn test_client_verifier_allow_unknown() {
         // We should be able to build a client verifier that allows unknown revocation status
-        let builder = WebPkiClientVerifier::builder(test_roots())
-            .with_crls(test_crls())
-            .allow_unknown_revocation_status();
+        let builder = WebPkiClientVerifier::builder_with_provider(
+            test_roots(),
+            provider::default_provider().into(),
+        )
+        .with_crls(test_crls())
+        .allow_unknown_revocation_status();
         // The builder should be Debug.
         println!("{:?}", builder);
         builder.build().unwrap();
@@ -566,7 +608,11 @@ mod tests {
     #[test]
     fn test_builder_no_roots() {
         // Trying to create a client verifier builder with no trust anchors should fail at build time
-        let result = WebPkiClientVerifier::builder(RootCertStore::empty().into()).build();
+        let result = WebPkiClientVerifier::builder_with_provider(
+            RootCertStore::empty().into(),
+            provider::default_provider().into(),
+        )
+        .build();
         assert!(matches!(result, Err(VerifierBuilderError::NoRootAnchors)));
     }
 
